
// Copyright (C) Tom Horsley, 1997. All rights reserved.

#include "ntp-long-fixed.h"
#include "ms-to-ntp.h"

// Generally speaking, the julian day number used by astronomers is the most
// convenient intermediate form to go through for calendar calculations. The
// JulianDayNumber and GregorianDate routines are used to do that conversion.

inline long int
JulianDayNumber(WORD wYear, WORD wMonth, WORD wDay)
{
   long int y = (long int)wYear;
   long int d = (long int)wDay;
   long int m = (long int)wMonth;
   if (m > 2) {
      m = m - 3;
   } else {
      m = m + 9;
      y = y - 1;
   }
   long int c = y / 100;
   long int ya = y - 100 * c;
   long int j = (146097L * c) / 4 + (1461L * ya) / 4 + (153L * m + 2) / 5 +
                d + 1721119L;
   return j;
}

inline void
GregorianDate(
   long julian_time,
   WORD & wYear,
   WORD & wMonth,
   WORD & wDay)
{
   long j = julian_time - 1721119;
   long y = (4 * j - 1) / 146097;
   j = 4 * j - 1 - 146097 * y;
   long d = j / 4;
   j = (4 * d + 3) / 1461;
   d = 4 * d + 3 - 1461 * j;
   d = (d + 4) / 4;
   long m = (5 * d - 3) / 153;
   d = 5 * d - 3 - 153 * m;
   d = (d + 5) / 5;
   y = 100 * y + j;
   if (m < 10) {
      m = m + 3;
   } else {
      m = m - 9;
      y = y + 1;
   }
   wYear = (WORD)y;
   wMonth = (WORD)m;
   wDay = (WORD)d;
}

// The base date for NTP time is Jan 1, 1900. This is the julian rep for
// that date. (Some time around 2036 the NTP date with overflow. Once that
// day is over, simply plugging in a different base time should suffice to
// fix everything for another 136 years - getting it to work correctly
// during that day is more trouble than I feel like going through now :-).
//
#define JAN_1_1900 2415021L /* pre-computed JulianDayNumber(1900,1,1) */

NtpLongFixed::NtpLongFixed(const SYSTEMTIME & sys_time)
{
   long int j;

   // First compute the (astronomical) julian day number
   j = JulianDayNumber(sys_time.wYear, sys_time.wMonth, sys_time.wDay);

   // Convert to NTP days
   j -= JAN_1_1900;

   // Hours
   j = (j * 24L) + (long int)sys_time.wHour;

   // Minutes
   j = (j * 60L) + (long int)sys_time.wMinute;

   // Seconds
   j = (j * 60L) + (long int)sys_time.wSecond;

   integer_part = (unsigned long)j;
   fractional_part = MsToNtp::Convert((int)sys_time.wMilliseconds);
}

#define P1 ((double)2.0)
#define P2 (P1*P1)
#define P4 (P2*P2)
#define P8 (P4*P4)
#define P16 (P8*P8)
#define P32 (P16*P16)
#define CVT (((double)1000.0)/P32)
#define HALF_MSEC_FRAC 0x0020c49b

void
NtpLongFixed::ConvertToSYSTEMTIME(SYSTEMTIME & sys_time) const
{
   // Round number up by adding in 1/2 millisecond before we start

   NtpLongFixed rounded(*this);
   NtpLongFixed half_msec(0, HALF_MSEC_FRAC);
   rounded = rounded + half_msec;

   unsigned long ip = rounded.IntegerPart();

   // Get seconds
   sys_time.wSecond = (WORD)(ip % 60L);
   ip /= 60L;

   // Get minutes
   sys_time.wMinute = (WORD)(ip % 60L);
   ip /= 60L;

   // Get hours
   sys_time.wHour = (WORD)(ip % 24L);
   ip /= 24L;

   // Convert to julian day number
   long int j = ((long int)ip) + JAN_1_1900;

   // Get weekday info
   sys_time.wDayOfWeek = (WORD)((j + 1) % 7L);

   // Get month, day, year
   GregorianDate(j, sys_time.wYear, sys_time.wMonth, sys_time.wDay);

   // convert fraction to milliseconds
   sys_time.wMilliseconds = (WORD)(((double)rounded.FractionalPart()) * CVT);
}
