#ifndef ntp_long_fixed_h
#define ntp_long_fixed_h

// Copyright (C) Tom Horsley, 1997. All rights reserved.

#include <windows.h>

// NetworkOrderLongFixed is a plain struct holding the NTP timestamp value
// in network order.
//
struct NetworkOrderLongFixed {
   unsigned long integer_part;
   unsigned long fractional_part;
};

// NtpLongFixed is a class which implements various operations on NTP format
// 64 bit fixed point numbers.
//
class NtpLongFixed {
public:

   NtpLongFixed();

   NtpLongFixed(const NtpLongFixed & cpy);

   NtpLongFixed(const SYSTEMTIME & sys_time);

   NtpLongFixed(const NetworkOrderLongFixed & netlong);

   NtpLongFixed(unsigned long int_part, unsigned long fract_part);

   NtpLongFixed &
   operator = (const NtpLongFixed & newlf);

   unsigned long
   IntegerPart() const;

   unsigned long
   FractionalPart() const;

   void
   ConvertToNetworkOrder(NetworkOrderLongFixed & netlong) const;

   void
   ConvertToSYSTEMTIME(SYSTEMTIME & sys_time) const;

   NtpLongFixed
   UnsignedDivByTwo() const;

   NtpLongFixed
   SignedDivByTwo() const;

private:

   unsigned long integer_part;
   unsigned long fractional_part;
};

inline
NtpLongFixed::NtpLongFixed()
{
   integer_part = 0;
   fractional_part = 0;
}

inline
NtpLongFixed::NtpLongFixed(const NtpLongFixed & cpy)
{
   integer_part = cpy.integer_part;
   fractional_part = cpy.fractional_part;
}

inline
NtpLongFixed::NtpLongFixed(const NetworkOrderLongFixed & netlong)
{
   integer_part = ntohl(netlong.integer_part);
   fractional_part = ntohl(netlong.fractional_part);
}

inline
NtpLongFixed::NtpLongFixed(unsigned long int_part, unsigned long fract_part)
{
   integer_part = int_part;
   fractional_part = fract_part;
}

inline NtpLongFixed &
NtpLongFixed::operator =(const NtpLongFixed & newlf)
{
   integer_part = newlf.integer_part;
   fractional_part = newlf.fractional_part;
   return *this;
}

inline unsigned long
NtpLongFixed::IntegerPart() const
{
   return integer_part;
}

inline unsigned long
NtpLongFixed::FractionalPart() const
{
   return fractional_part;
}

inline void
NtpLongFixed::ConvertToNetworkOrder(NetworkOrderLongFixed & netlong) const
{
   netlong.integer_part = htonl(integer_part);
   netlong.fractional_part = htonl(fractional_part);
}

inline NtpLongFixed
operator + (const NtpLongFixed & arg1, const NtpLongFixed & arg2)
{
   unsigned long f1 = arg1.FractionalPart();
   unsigned long f2 = arg2.FractionalPart();
   unsigned long lo16 = (f1 & 0xffff) + (f2 & 0xffff);
   unsigned long hi16 = (f1 >> 16) + (f2 >> 16) + (lo16 >> 16);
   return NtpLongFixed(arg1.IntegerPart() + arg2.IntegerPart() + (hi16 >> 16),
                       ((hi16 << 16) | (lo16 & 0xffff)));
}

inline NtpLongFixed
operator ~ (const NtpLongFixed & arg)
{
   return NtpLongFixed(~arg.IntegerPart(), ~arg.FractionalPart());
}

inline NtpLongFixed
operator - (const NtpLongFixed & arg)
{
   return (~ arg) + NtpLongFixed(0,1);
}

inline NtpLongFixed
operator - (const NtpLongFixed & arg1, const NtpLongFixed & arg2)
{
   return arg1 + (- arg2);
}

inline NtpLongFixed
NtpLongFixed::UnsignedDivByTwo() const
{
   unsigned long ip = IntegerPart();
   unsigned long fp = FractionalPart();
   fp = ((fp >> 1) | (ip << 31));
   ip = ip >> 1;
   return NtpLongFixed(ip, fp);
}

inline NtpLongFixed
NtpLongFixed::SignedDivByTwo() const
{
   if (IntegerPart() & 0x80000000) {
      return -((- (* this)).UnsignedDivByTwo());
   } else {
      return UnsignedDivByTwo();
   }
}

#endif // ntp_long_fixed_h
