#include "ntp-real-events.h"
#include "lasterr-exception.h"
#include "simple-exception.h"
#include "events.h"
#include "log-stuff.h"
#include <string.h>

NtpRealEvents::NtpRealEvents(bool is_nt,
                             bool allow_poke,
                             bool allow_monitor,
                             LogStuff * logp) {

   // Ack! To create a named event anyone can reference, I have to dive into
   // the security API morass, but since the interfaces may or may not exist
   // in different versions of Windows, I also need to dyanmically look them
   // up to decide if I should call them.
   //
   // For the "poke" event, I want anyone to be able to set it.
   //
   // For the state events, I'm the only one who should be able to set and
   // reset, but anyone should be able to monitor them. (Experimentation
   // reveals that I have access to set them simply because I create them,
   // so the only actual access I need to set is the one to allow everyone
   // else to sync on them).

   typedef BOOL WINAPI
   AllocateAndInitializeSid_func(PSID_IDENTIFIER_AUTHORITY,BYTE,DWORD,DWORD,
                                 DWORD,DWORD,DWORD,DWORD,DWORD,DWORD,PSID*);

   typedef BOOL WINAPI
   InitializeAcl_func(PACL,DWORD,DWORD);

   typedef BOOL WINAPI
   AddAccessAllowedAce_func(PACL,DWORD,DWORD,PSID);

   typedef BOOL WINAPI
   InitializeSecurityDescriptor_func(PSECURITY_DESCRIPTOR,DWORD);

   typedef BOOL WINAPI
   SetSecurityDescriptorDacl_func(PSECURITY_DESCRIPTOR,BOOL,PACL,BOOL);

   typedef DWORD WINAPI
   GetLengthSid_func(PSID);

   typedef PVOID WINAPI
   FreeSid_func(PSID);

   SID_IDENTIFIER_AUTHORITY siaWorld = SECURITY_WORLD_SID_AUTHORITY;
   PSID psidEveryone = NULL;
   PACL acl_sync = NULL;
   PACL acl_all = NULL;
   SECURITY_DESCRIPTOR sd_sync;
   SECURITY_DESCRIPTOR sd_all;
   SECURITY_ATTRIBUTES sa_sync;
   SECURITY_ATTRIBUTES sa_all;

   ZeroMemory((void *)&sa_sync, sizeof(sa_sync));
   sa_sync.nLength = sizeof(sa_sync);
   sa_sync.bInheritHandle = FALSE;
   sa_sync.lpSecurityDescriptor = NULL;

   ZeroMemory((void *)&sa_all, sizeof(sa_all));
   sa_all.nLength = sizeof(sa_all);
   sa_all.bInheritHandle = FALSE;
   sa_all.lpSecurityDescriptor = NULL;

   is_valid = true;

   poke_event = NULL;
   for (int i = 0; i < sizeof(states)/sizeof(states[0]); ++i) {
      states[i] = NULL;
   }

   HINSTANCE advapi_dll = NULL;
   AllocateAndInitializeSid_func     * aaisf = NULL;
   InitializeAcl_func                * iaf   = NULL;
   AddAccessAllowedAce_func          * aaaaf = NULL;
   InitializeSecurityDescriptor_func * isdf  = NULL;
   SetSecurityDescriptorDacl_func    * ssddf = NULL;
   GetLengthSid_func                 * glsf  = NULL;
   FreeSid_func                      * fsf   = NULL;

   try {

      // Get handle to security lib and lookup the function pointers I need
      // to create the proper access lists for the event objects.

      advapi_dll = LoadLibrary("Advapi32");
      if (advapi_dll == NULL) {
         throw new LastErrorException("LoadLibrary(Advapi32)");
      }
      aaisf = (AllocateAndInitializeSid_func *)
         GetProcAddress(advapi_dll, "AllocateAndInitializeSid");
      if (aaisf == NULL) {
         throw new SimpleException("lookup failed: AllocateAndInitializeSid");
      }
      iaf = (InitializeAcl_func *)
         GetProcAddress(advapi_dll, "InitializeAcl");
      if (iaf == NULL) {
         throw new SimpleException("lookup failed: InitializeAcl");
      }
      aaaaf = (AddAccessAllowedAce_func *)
         GetProcAddress(advapi_dll, "AddAccessAllowedAce");
      if (aaaaf == NULL) {
         throw new SimpleException("lookup failed: AddAccessAllowedAce");
      }
      isdf = (InitializeSecurityDescriptor_func *)
         GetProcAddress(advapi_dll, "InitializeSecurityDescriptor");
      if (isdf == NULL) {
         throw new SimpleException("lookup failed: InitializeSecurityDescriptor");
      }
      ssddf = (SetSecurityDescriptorDacl_func *)
         GetProcAddress(advapi_dll, "SetSecurityDescriptorDacl");
      if (ssddf == NULL) {
         throw new SimpleException("lookup failed: SetSecurityDescriptorDacl");
      }
      glsf = (GetLengthSid_func *)
         GetProcAddress(advapi_dll, "GetLengthSid");
      if (glsf == NULL) {
         throw new SimpleException("lookup failed: GetLengthSid");
      }
      fsf = (FreeSid_func *)
         GetProcAddress(advapi_dll, "FreeSid");
      if (fsf == NULL) {
         throw new SimpleException("lookup failed: FreeSid");
      }

      // Create the SID for "Everyone"

      if (!(*aaisf)(&siaWorld, 1, SECURITY_WORLD_RID, 0, 0, 0, 0, 0, 0, 0,
                    &psidEveryone)) {
         psidEveryone = NULL;
         throw new LastErrorException("AllocateAndInitializeSid(everyone)");
      }

      // Now make an ACL and SD that says everyone can sync

      DWORD acl_size = sizeof(ACL) +
                       sizeof (ACCESS_ALLOWED_ACE) -
                       sizeof (DWORD) +
                       (*glsf)(psidEveryone);
      acl_sync = (PACL)(LocalAlloc(LPTR, acl_size));
      if (! acl_sync) {
         throw new LastErrorException("LocalAlloc()");
      }
      if (! (*iaf)(acl_sync, acl_size, ACL_REVISION)) {
         throw new LastErrorException("InitializeAcl(sync)");
      }
      if(!(*aaaaf)(acl_sync, ACL_REVISION, SYNCHRONIZE|GENERIC_READ,
                   psidEveryone)) {
         throw new LastErrorException("AddAccessAllowedAce(sync)");
      }
      if (! (*isdf)(&sd_sync, SECURITY_DESCRIPTOR_REVISION)) {
         throw new LastErrorException("InitializeSecurityDescriptor(sync)");
      }
      if (! (*ssddf)(&sd_sync, TRUE, acl_sync, FALSE)) {
         throw new LastErrorException("SetSecurityDescriptorDacl(sync)");
      }

      // And another ACL and SD that says everyone can do anything

      acl_all = (PACL)(LocalAlloc(LPTR, acl_size));
      if (! acl_all) {
         throw new LastErrorException("LocalAlloc()");
      }
      if (! (*iaf)(acl_all, acl_size, ACL_REVISION)) {
         throw new LastErrorException("InitializeAcl(all)");
      }
      if(!(*aaaaf)(acl_all, ACL_REVISION, EVENT_ALL_ACCESS,
                   psidEveryone)) {
         throw new LastErrorException("AddAccessAllowedAce(all)");
      }
      if (! (*isdf)(&sd_all, SECURITY_DESCRIPTOR_REVISION)) {
         throw new LastErrorException("InitializeSecurityDescriptor(all)");
      }
      if (! (*ssddf)(&sd_all, TRUE, acl_all, FALSE)) {
         throw new LastErrorException("SetSecurityDescriptorDacl(all)");
      }

      // If we made it this far, we have valid security descriptors
      // so set the pointers in the security attributes.

      sa_sync.lpSecurityDescriptor = &sd_sync;
      sa_all.lpSecurityDescriptor = &sd_all;
   } catch (std::exception * ee) {
      if (is_nt) {

         // If we were running on NT/2000/XP, we expect this stuff to have
         // worked. If it doesn't, the log entry is appropriate.

         logp->LogPrintf("event security descriptors failure: %s", ee->what());
         is_valid = false;
      }
      delete ee;
   }

   if (is_valid && allow_poke) {
      poke_event = MakeGlobalEvent(&sa_all, TRUE, FALSE, EV_NTPTIME_POKE);
      if (poke_event == NULL) is_valid = false;
   }

   // Create states[EV_INIT] in set state, others in reset state.

   if (is_valid && allow_monitor) {
      states[EV_INIT] = MakeGlobalEvent(&sa_sync, TRUE, TRUE, EV_NTPTIME_INIT);
      cur_state = EV_INIT;
      if (states[EV_INIT] == NULL) is_valid = false;
      states[EV_OK] = MakeGlobalEvent(&sa_sync, TRUE, FALSE, EV_NTPTIME_OK);
      if (states[EV_OK] == NULL) is_valid = false;
      states[EV_WARN] = MakeGlobalEvent(&sa_sync, TRUE, FALSE, EV_NTPTIME_WARN);
      if (states[EV_WARN] == NULL) is_valid = false;
      states[EV_CLOSE] = MakeGlobalEvent(&sa_sync, TRUE, FALSE, EV_NTPTIME_CLOSE);
      if (states[EV_CLOSE] == NULL) is_valid = false;
   }

   // Clean up any storage that may or may not have been allocated above

   if (acl_sync != NULL) LocalFree(acl_sync);
   if (acl_all != NULL) LocalFree(acl_all);
   if ((psidEveryone != NULL) && (fsf != NULL)) (*fsf)(psidEveryone);
   if (advapi_dll != NULL) FreeLibrary(advapi_dll);
}

NtpRealEvents::~NtpRealEvents() {
   CloseHandle(poke_event);
   for (int i = 0; i < sizeof(states)/sizeof(states[0]); ++i) {
      CloseHandle(states[i]);
   }
}

bool
NtpRealEvents::IsValid() {
   return is_valid;
}

DWORD
NtpRealEvents::AppendExtraEvents(HANDLE ev[], DWORD evcount) {
   if (poke_event != NULL) {
      ev[evcount++] = poke_event;
   }
   return evcount;
}

void
NtpRealEvents::ResetPoke() {
   ResetEvent(poke_event);
}

void
NtpRealEvents::SetInitState() {
   CommonStateChange(EV_INIT);
}

void
NtpRealEvents::SetSyncOK() {
   CommonStateChange(EV_OK);
}

void
NtpRealEvents::SetSyncWarn() {
   CommonStateChange(EV_WARN);
}

void
NtpRealEvents::SetShutdownState() {
   CommonStateChange(EV_CLOSE);
}

void
NtpRealEvents::CommonStateChange(ev_state set_me) {
   if (cur_state != set_me) {
      if (states[cur_state] != NULL) ResetEvent(states[cur_state]);
      cur_state = set_me;
      if (states[cur_state] != NULL) SetEvent(states[cur_state]);
   }
}

HANDLE
NtpRealEvents::MakeGlobalEvent(
   LPSECURITY_ATTRIBUTES sap,
   BOOL manual_reset,
   BOOL initial_state,
   LPCTSTR name) {

   char globname[100];

   if ((strlen(name) + 8) >= sizeof(globname)) {
      return NULL;
   }

   // First try the name with the Global\ prefix, if that doesn't
   // work on this version of windows, try again without the Global\

   strcpy(globname, "Global\\");
   strcat(globname, name);
   HANDLE rval = CreateEvent(sap, manual_reset, initial_state, globname);
   if (rval == NULL) {
      rval = CreateEvent(sap, manual_reset, initial_state, name);
   }
   return rval;
}
