#ifndef ftrace_h
#define ftrace_h

#pragma warning(disable:4786)

// Function tracing and exceptions.

// Here's the rules:
//
// Always start every function with an FTBEG line and end it with FTEND
// FTBEG should get a printf-like argument list as its single argument.
//
// Example:
//
// int
// factorial(int x) {
//    FTBEG(("factorial(%d)", x))
//    if (x <= 1) {
//       return 1;
//    } else {
//       return factorial(x-1) * x;
//    }
//    FTEND
// }
//
// If you throw any exception (based on the BaseException) class,
// do it like this:
//
//   FTHROW(new LastErrorException(GetLastError(), "CreateFile()"));
//
// What all this accomplishes is adding the callback trace info to
// the exception as it proagates up the stack to the routine which
// finally wants to really handle it. It also adds the file:line
// info from the location of the FTHROW.

#ifndef ACTIVATE_FTRACE
#define ACTIVATE_FTRACE 1
#endif

#if ACTIVATE_FTRACE >= 1

#include <string>
#include "base-exception.h"

class FunctionTracer {
public:

   FunctionTracer(const char * fmt, ...);

   ~FunctionTracer();

   const std::string &
   GetFunctionTraceString() const;

private:

   std::string function_trace_string;
};

#define FTBEG(args) \
\
   FunctionTracer trace__me args; \
   try {

#define FTEND \
 \
   } catch (BaseException * e) { \
      e->AddCalledFromInfo(trace__me); \
      throw e; \
   }

#define FTHROW(e) \
\
   { \
      BaseException * e__ = (e ); \
      e__->AddThrowLocation(__FILE__, __LINE__); \
      throw e__; \
   }

#else

#define FTBEG(args)

#define FTEND

#define FTHROW(e) throw e

#endif

#endif // ftrace_h
