#ifndef wincmd_h
#define wincmd_h

// Copyright (C) Tom Horsley, 1998. All rights reserved.

#include <windows.h>
#include <string>
#include <vector>

// We can recognize three different kinds of arguments - string, long
// integer, and bool. Anything else, you need to pick up as string and
// handle yourself.
//
enum ArgKind {
   WINCMD_STRING,
   WINCMD_LONG,
   WINCMD_BOOL
};

// We can find argument values 3 different ways - we use the default built
// into the program because the argument wasn't specified, we find an explicit
// argument on the command line, or we find the argument value via a registry
// lookup.
//
enum ArgOrigin {
   WINCMD_UNSPECIFIED,
   WINCMD_COMMAND_LINE,
   WINCMD_REGISTRY
};

// An argument has a name, a kind, and a description (for help functions).
// It may also have a default value which can be obtained from the registry.
// The arg_val_ptr points to the variable (of type char *, long, or bool)
// which will receive the value of the argument when parsed by
// InterpretArgs.
//
struct ArgDef {
   const char * arg_name;
   ArgKind      arg_kind;
   const char * arg_description;
   HKEY         arg_which_registry;
   const char * arg_registry_key;
   const char * arg_registry_value;
   DWORD        arg_value_type;
   void *       arg_val_ptr;
   ArgOrigin    arg_origin;
};

// Parse a command line string into a bunch of argument strings (no
// wildcards, just space separated strings with special processing for quote
// and backslash-quote).
//
class Wincmd {
public:

   Wincmd(PSTR szCmdLine);

   ~Wincmd();

   int
   Argc();

   std::string &
   operator[](int i);

      // Go through the argument array, finding matching arguments and
      // filling in the values via the arg_val_ptr. argdefs is a pointer to
      // an array of ArgDef structs.
      //
      // Following this call, any arguments matched will be removed from the
      // array and only positional parameters will remain in the list.
      //
      // Throws a std::string with any error message should something
      // go wrong.
      //
   void
   InterpretArgs(ArgDef * argdef, int argdef_count);

      // Go through the arg def info and append descriptions of all the
      // options to the msg string.
      //
   static void
   AppendHelpInfo(std::string & msg, ArgDef * argdef, int argdef_count);

private:

   std::vector<std::string>  argv;

   int skip;
};

#endif // wincmd_h
